@@ -2,7 +2,7 @@
  *  Copyright 2004-2005 Paul Querna
  *  Copyright 2008, 2014 Nikos Mavrogiannopoulos
  *  Copyright 2011 Dash Shendy
- *  Copyright 2015-2016 Fiona Klute
+ *  Copyright 2015-2020 Fiona Klute
  *
  *  Licensed under the Apache License, Version 2.0 (the "License");
  *  you may not use this file except in compliance with the License.
@@ -17,10 +17,13 @@
  *  limitations under the License.
  */
 
+#include "gnutls_cache.h"
 #include "gnutls_config.h"
 #include "mod_gnutls.h"
 #include "gnutls_ocsp.h"
+
 #include "apr_lib.h"
+#include <apr_strings.h>
 #include <gnutls/abstract.h>
 
 #define INIT_CA_SIZE 128
@@ -141,31 +144,10 @@ static apr_status_t mgs_pool_free_credentials(void *arg)
         sc->ca_list = NULL;
     }
 
-    if (sc->cert_pgp)
-    {
-        gnutls_pcert_deinit(&sc->cert_pgp[0]);
-        sc->cert_pgp = NULL;
-        gnutls_openpgp_crt_deinit(sc->cert_crt_pgp[0]);
-        sc->cert_crt_pgp = NULL;
-    }
-
-    if (sc->privkey_pgp)
-    {
-        gnutls_privkey_deinit(sc->privkey_pgp);
-        sc->privkey_pgp = NULL;
-#if GNUTLS_VERSION_NUMBER < 0x030312
-        gnutls_openpgp_privkey_deinit(sc->privkey_pgp_internal);
-        sc->privkey_pgp_internal = NULL;
-#endif
-    }
-
-    if (sc->pgp_list)
-    {
-        gnutls_openpgp_keyring_deinit(sc->pgp_list);
-        sc->pgp_list = NULL;
-    }
-
-    if (sc->priorities)
+    /* Deinit server priorities only if set from
+     * sc->priorities_str. Otherwise the server is using the default
+     * global priority cache, which must not be deinitialized here. */
+    if (sc->priorities_str && sc->priorities)
     {
         gnutls_priority_deinit(sc->priorities);
         sc->priorities = NULL;
@@ -452,161 +434,6 @@ int mgs_load_files(apr_pool_t *pconf, apr_pool_t *ptemp, server_rec *s)
         }
     }
 
-    if (sc->pgp_cert_file && sc->cert_pgp == NULL)
-    {
-        sc->cert_pgp = apr_pcalloc(pconf, sizeof(sc->cert_pgp[0]));
-        sc->cert_crt_pgp = apr_pcalloc(pconf, sizeof(sc->cert_crt_pgp[0]));
-
-        if (load_datum_from_file(spool, sc->pgp_cert_file, &data) != 0) {
-            ap_log_error(APLOG_MARK, APLOG_STARTUP, 0, s,
-                         "GnuTLS: Error Reading " "Certificate '%s'",
-                         sc->pgp_cert_file);
-            ret = -1;
-            goto cleanup;
-        }
-
-        ret = gnutls_openpgp_crt_init(&sc->cert_crt_pgp[0]);
-        if (ret < 0) {
-            ap_log_error(APLOG_MARK, APLOG_STARTUP, 0, s,
-                         "GnuTLS: Failed to Init "
-                         "PGP Certificate: (%d) %s", ret,
-                         gnutls_strerror(ret));
-            ret = -1;
-            goto cleanup;
-        }
-
-        ret = gnutls_openpgp_crt_import(sc->cert_crt_pgp[0], &data,
-                                        GNUTLS_OPENPGP_FMT_BASE64);
-        if (ret < 0) {
-            ap_log_error(APLOG_MARK, APLOG_STARTUP, 0, s,
-                         "GnuTLS: Failed to Import "
-                         "PGP Certificate: (%d) %s", ret,
-                         gnutls_strerror(ret));
-            ret = -1;
-            goto cleanup;
-        }
-
-        ret = gnutls_pcert_import_openpgp(sc->cert_pgp,
-                                          sc->cert_crt_pgp[0], 0);
-        if (ret < 0) {
-            ap_log_error(APLOG_MARK, APLOG_STARTUP, 0, s,
-                         "GnuTLS: Failed to Import "
-                         "PGP pCertificate: (%d) %s", ret,
-                         gnutls_strerror(ret));
-            ret = -1;
-            goto cleanup;
-        }
-    }
-
-    /* Load the PGP key file */
-    if (sc->pgp_key_file && sc->privkey_pgp == NULL)
-    {
-        if (load_datum_from_file(spool, sc->pgp_key_file, &data) != 0) {
-            ap_log_error(APLOG_MARK, APLOG_STARTUP, 0, s,
-                         "GnuTLS: Error Reading " "Private Key '%s'",
-                         sc->pgp_key_file);
-            ret = -1;
-            goto cleanup;
-        }
-
-        ret = gnutls_privkey_init(&sc->privkey_pgp);
-        if (ret < 0) {
-            ap_log_error(APLOG_MARK, APLOG_STARTUP, 0, s,
-                         "GnuTLS: Failed to initialize"
-                         ": (%d) %s", ret, gnutls_strerror(ret));
-            ret = -1;
-            goto cleanup;
-        }
-
-#if GNUTLS_VERSION_NUMBER < 0x030312
-        /* GnuTLS versions before 3.3.12 contain a bug in
-         * gnutls_privkey_import_openpgp_raw which frees data that is
-         * accessed when the key is used, leading to segfault. Loading
-         * the key into a gnutls_openpgp_privkey_t and then assigning
-         * it to the gnutls_privkey_t works around the bug, hence this
-         * chain of gnutls_openpgp_privkey_init,
-         * gnutls_openpgp_privkey_import and
-         * gnutls_privkey_import_openpgp. */
-        ret = gnutls_openpgp_privkey_init(&sc->privkey_pgp_internal);
-        if (ret != 0) {
-            ap_log_error(APLOG_MARK, APLOG_STARTUP, 0, s,
-                         "GnuTLS: Failed to initialize "
-                         "PGP Private Key '%s': (%d) %s",
-                         sc->pgp_key_file, ret, gnutls_strerror(ret));
-            ret = -1;
-            goto cleanup;
-        }
-
-        ret = gnutls_openpgp_privkey_import(sc->privkey_pgp_internal, &data,
-                                            GNUTLS_OPENPGP_FMT_BASE64, NULL, 0);
-        if (ret != 0) {
-            ap_log_error(APLOG_MARK, APLOG_STARTUP, 0, s,
-                         "GnuTLS: Failed to Import "
-                         "PGP Private Key '%s': (%d) %s",
-                         sc->pgp_key_file, ret, gnutls_strerror(ret));
-            ret = -1;
-            goto cleanup;
-        }
-
-        ret = gnutls_privkey_import_openpgp(sc->privkey_pgp,
-                                            sc->privkey_pgp_internal, 0);
-        if (ret != 0)
-        {
-            ap_log_error(APLOG_MARK, APLOG_STARTUP, 0, s,
-                         "GnuTLS: Failed to assign PGP Private Key '%s' "
-                         "to gnutls_privkey_t structure: (%d) %s",
-                         sc->pgp_key_file, ret, gnutls_strerror(ret));
-            ret = -1;
-            goto cleanup;
-        }
-#else
-        ret = gnutls_privkey_import_openpgp_raw(sc->privkey_pgp, &data,
-                                                GNUTLS_OPENPGP_FMT_BASE64,
-                                                NULL, NULL);
-        if (ret != 0)
-        {
-            ap_log_error(APLOG_MARK, APLOG_STARTUP, 0, s,
-                         "GnuTLS: Failed to Import "
-                         "PGP Private Key '%s': (%d) %s",
-                         sc->pgp_key_file, ret, gnutls_strerror(ret));
-            ret = -1;
-            goto cleanup;
-        }
-#endif
-    }
-
-    /* Load the keyring file */
-    if (sc->pgp_ring_file && sc->pgp_list == NULL)
-    {
-        if (load_datum_from_file(spool, sc->pgp_ring_file, &data) != 0) {
-            ap_log_error(APLOG_MARK, APLOG_STARTUP, 0, s,
-                         "GnuTLS: Error Reading " "Keyring File '%s'",
-                         sc->pgp_ring_file);
-            ret = -1;
-            goto cleanup;
-        }
-
-        ret = gnutls_openpgp_keyring_init(&sc->pgp_list);
-        if (ret < 0) {
-            ap_log_error(APLOG_MARK, APLOG_STARTUP, 0, s,
-                         "GnuTLS: Failed to initialize"
-                         "keyring: (%d) %s", ret, gnutls_strerror(ret));
-            ret = -1;
-            goto cleanup;
-        }
-
-        ret = gnutls_openpgp_keyring_import(sc->pgp_list, &data,
-                                            GNUTLS_OPENPGP_FMT_BASE64);
-        if (ret < 0) {
-            ap_log_error(APLOG_MARK, APLOG_STARTUP, 0, s,
-                         "GnuTLS: Failed to load "
-                         "Keyring File '%s': (%d) %s", sc->pgp_ring_file,
-                         ret, gnutls_strerror(ret));
-            ret = -1;
-            goto cleanup;
-        }
-    }
-
     if (sc->priorities_str && sc->priorities == NULL)
     {
         const char *err;
@@ -714,31 +541,6 @@ const char *mgs_set_key_file(cmd_parms * parms, void *dummy __attribute__((unuse
     return NULL;
 }
 
-const char *mgs_set_pgpcert_file(cmd_parms * parms, void *dummy __attribute__((unused)),
-        const char *arg)
-{
-    mgs_srvconf_rec *sc =
-	(mgs_srvconf_rec *) ap_get_module_config(parms->server->
-						 module_config,
-						 &gnutls_module);
-
-    sc->pgp_cert_file = ap_server_root_relative(parms->pool, arg);
-
-    return NULL;
-}
-
-const char *mgs_set_pgpkey_file(cmd_parms * parms, void *dummy __attribute__((unused)),
-        const char *arg) {
-    mgs_srvconf_rec *sc =
-	(mgs_srvconf_rec *) ap_get_module_config(parms->server->
-						 module_config,
-						 &gnutls_module);
-
-    sc->pgp_key_file = ap_server_root_relative(parms->pool, arg);
-
-    return NULL;
-}
-
 const char *mgs_set_tickets(cmd_parms *parms,
                             void *dummy __attribute__((unused)),
                             const int arg)
@@ -783,45 +585,64 @@ const char *mgs_set_srp_tpasswd_conf_file(cmd_parms * parms, void *dummy __attri
 
 #endif
 
-const char *mgs_set_cache(cmd_parms * parms, void *dummy __attribute__((unused)),
-        const char *type, const char *arg) {
+const char *mgs_set_cache(cmd_parms * parms,
+                          void *dummy __attribute__((unused)),
+                          const char *type, const char *arg)
+{
     const char *err;
     mgs_srvconf_rec *sc =
-	ap_get_module_config(parms->server->module_config,
-			     &gnutls_module);
-    if ((err = ap_check_cmd_context(parms, GLOBAL_ONLY))) {
-	return err;
+        ap_get_module_config(parms->server->module_config,
+                             &gnutls_module);
+    if ((err = ap_check_cmd_context(parms, GLOBAL_ONLY)))
+        return err;
+
+    unsigned char enable = GNUTLS_ENABLED_TRUE;
+    /* none: disable cache */
+    if (strcasecmp("none", type) == 0)
+        enable = GNUTLS_ENABLED_FALSE;
+
+    /* Try to split socache "type:config" style configuration */
+    const char* sep = ap_strchr_c(type, ':');
+    if (sep)
+    {
+        type = apr_pstrmemdup(parms->temp_pool, type, sep - type);
+        if (arg != NULL)
+        {
+            /* No mixing of socache style and legacy config! */
+            return "GnuTLSCache appears to have a mod_socache style "
+                "type:config value, but there is a second parameter!";
+        }
+        arg = ++sep;
     }
 
-    if (strcasecmp("none", type) == 0) {
-	sc->cache_type = mgs_cache_none;
-	sc->cache_config = NULL;
-	return NULL;
-    } else if (strcasecmp("dbm", type) == 0) {
-	sc->cache_type = mgs_cache_dbm;
-    } else if (strcasecmp("gdbm", type) == 0) {
-	sc->cache_type = mgs_cache_gdbm;
+    mgs_cache_t *cache = NULL;
+    /* parms->directive->directive contains the directive string */
+    if (!strcasecmp(parms->directive->directive, "GnuTLSCache"))
+    {
+        if (enable == GNUTLS_ENABLED_FALSE)
+        {
+            sc->cache_enable = GNUTLS_ENABLED_FALSE;
+            return NULL;
+        }
+        sc->cache_enable = GNUTLS_ENABLED_TRUE;
+        cache = &sc->cache;
     }
-#if HAVE_APR_MEMCACHE
-    else if (strcasecmp("memcache", type) == 0) {
-	sc->cache_type = mgs_cache_memcache;
-    }
-#endif
-    else {
-	return "Invalid Type for GnuTLSCache!";
+    else if (!strcasecmp(parms->directive->directive, "GnuTLSOCSPCache"))
+    {
+        if (enable == GNUTLS_ENABLED_FALSE)
+            return "\"GnuTLSOCSPCache none\" is invalid, use "
+                "\"GnuTLSOCSPStapling off\" if you want to disable "
+                "OCSP stapling.";
+        cache = &sc->ocsp_cache;
     }
+    else
+        return apr_psprintf(parms->temp_pool, "Internal Error: %s "
+                            "called for unknown directive %s",
+                            __func__, parms->directive->directive);
 
-    if (arg == NULL)
-	return "Invalid argument 2 for GnuTLSCache!";
-
-    if (sc->cache_type == mgs_cache_dbm
-	|| sc->cache_type == mgs_cache_gdbm) {
-	sc->cache_config = ap_server_root_relative(parms->pool, arg);
-    } else {
-	sc->cache_config = apr_pstrdup(parms->pool, arg);
-    }
-
-    return NULL;
+    return mgs_cache_inst_config(cache, parms->server,
+                                 type, arg,
+                                 parms->pool, parms->temp_pool);
 }
 
 const char *mgs_set_timeout(cmd_parms * parms,
@@ -837,21 +658,19 @@ const char *mgs_set_timeout(cmd_parms * parms,
     mgs_srvconf_rec *sc = (mgs_srvconf_rec *)
         ap_get_module_config(parms->server->module_config, &gnutls_module);
 
-    if (!apr_strnatcasecmp(parms->directive->directive, "GnuTLSCacheTimeout"))
-    {
-        const char *err;
-        if ((err = ap_check_cmd_context(parms, GLOBAL_ONLY)))
-            return err;
+    if (!strcasecmp(parms->directive->directive, "GnuTLSCacheTimeout"))
         sc->cache_timeout = apr_time_from_sec(argint);
-    }
-    else if (!apr_strnatcasecmp(parms->directive->directive,
-                                "GnuTLSOCSPCacheTimeout"))
+    else if (!strcasecmp(parms->directive->directive,
+                         "GnuTLSOCSPCacheTimeout"))
         sc->ocsp_cache_time = apr_time_from_sec(argint);
-    else if (!apr_strnatcasecmp(parms->directive->directive,
-                                "GnuTLSOCSPFailureTimeout"))
+    else if (!strcasecmp(parms->directive->directive,
+                         "GnuTLSOCSPFailureTimeout"))
         sc->ocsp_failure_timeout = apr_time_from_sec(argint);
-    else if (!apr_strnatcasecmp(parms->directive->directive,
-                                "GnuTLSOCSPSocketTimeout"))
+    else if (!strcasecmp(parms->directive->directive,
+                         "GnuTLSOCSPFuzzTime"))
+        sc->ocsp_fuzz_time = apr_time_from_sec(argint);
+    else if (!strcasecmp(parms->directive->directive,
+                         "GnuTLSOCSPSocketTimeout"))
         sc->ocsp_socket_timeout = apr_time_from_sec(argint);
     else
         /* Can't happen unless there's a serious bug in mod_gnutls or Apache */
@@ -923,18 +742,6 @@ const char *mgs_set_client_ca_file(cmd_parms * parms, void *dummy __attribute__(
     return NULL;
 }
 
-const char *mgs_set_keyring_file(cmd_parms * parms, void *dummy __attribute__((unused)),
-        const char *arg) {
-    mgs_srvconf_rec *sc =
-	(mgs_srvconf_rec *) ap_get_module_config(parms->server->
-						 module_config,
-						 &gnutls_module);
-
-    sc->pgp_ring_file = ap_server_root_relative(parms->pool, arg);
-
-    return NULL;
-}
-
 /*
  * Enable TLS proxy operation if arg is true, disable it otherwise.
  */
@@ -1070,18 +877,9 @@ static mgs_srvconf_rec *_mgs_config_server_create(apr_pool_t * p,
     sc->p11_modules = NULL;
     sc->pin = NULL;
 
-    sc->cert_pgp = NULL;
-    sc->cert_crt_pgp = NULL;
-    sc->privkey_pgp = NULL;
-#if GNUTLS_VERSION_NUMBER < 0x030312
-    sc->privkey_pgp_internal = NULL;
-#endif
-    sc->pgp_list = NULL;
-
     sc->priorities_str = NULL;
     sc->cache_timeout = MGS_TIMEOUT_UNSET;
-    sc->cache_type = mgs_cache_unset;
-    sc->cache_config = NULL;
+    sc->cache_enable = GNUTLS_ENABLED_UNSET;
     sc->cache = NULL;
     sc->tickets = GNUTLS_ENABLED_UNSET;
     sc->priorities = NULL;
@@ -1104,13 +902,21 @@ static mgs_srvconf_rec *_mgs_config_server_create(apr_pool_t * p,
     sc->proxy_x509_tl = NULL;
 
     sc->ocsp_staple = GNUTLS_ENABLED_UNSET;
+    sc->ocsp_auto_refresh = GNUTLS_ENABLED_UNSET;
     sc->ocsp_check_nonce = GNUTLS_ENABLED_UNSET;
     sc->ocsp_response_file = NULL;
+    sc->ocsp_response_file_num = 0;
+    sc->ocsp = NULL;
+    sc->ocsp_num = 0;
     sc->ocsp_mutex = NULL;
+    sc->ocsp_cache = NULL;
     sc->ocsp_cache_time = MGS_TIMEOUT_UNSET;
     sc->ocsp_failure_timeout = MGS_TIMEOUT_UNSET;
+    sc->ocsp_fuzz_time = MGS_TIMEOUT_UNSET;
     sc->ocsp_socket_timeout = MGS_TIMEOUT_UNSET;
 
+    sc->singleton_wd = NULL;
+
 /* this relies on GnuTLS never changing the gnutls_certificate_request_t enum to define -1 */
     sc->client_verify_mode = -1;
 
@@ -1132,7 +938,6 @@ void *mgs_config_server_create(apr_pool_t * p,
 
 void *mgs_config_server_merge(apr_pool_t * p, void *BASE, void *ADD)
 {
-    int i;
     char *err = NULL;
     mgs_srvconf_rec *base = (mgs_srvconf_rec *) BASE;
     mgs_srvconf_rec *add = (mgs_srvconf_rec *) ADD;
@@ -1154,11 +959,9 @@ void *mgs_config_server_merge(apr_pool_t * p, void *BASE, void *ADD)
     gnutls_srvconf_merge(x509_ca_file, NULL);
     gnutls_srvconf_merge(p11_modules, NULL);
     gnutls_srvconf_merge(pin, NULL);
-    gnutls_srvconf_merge(pgp_cert_file, NULL);
-    gnutls_srvconf_merge(pgp_key_file, NULL);
-    gnutls_srvconf_merge(pgp_ring_file, NULL);
     gnutls_srvconf_merge(dh_file, NULL);
     gnutls_srvconf_merge(priorities_str, NULL);
+    gnutls_srvconf_merge(cache_timeout, MGS_TIMEOUT_UNSET);
 
     gnutls_srvconf_merge(proxy_x509_key_file, NULL);
     gnutls_srvconf_merge(proxy_x509_cert_file, NULL);
@@ -1168,17 +971,17 @@ void *mgs_config_server_merge(apr_pool_t * p, void *BASE, void *ADD)
     gnutls_srvconf_merge(proxy_priorities, NULL);
 
     gnutls_srvconf_merge(ocsp_staple, GNUTLS_ENABLED_UNSET);
+    gnutls_srvconf_merge(ocsp_auto_refresh, GNUTLS_ENABLED_UNSET);
     gnutls_srvconf_merge(ocsp_check_nonce, GNUTLS_ENABLED_UNSET);
     gnutls_srvconf_assign(ocsp_response_file);
+    gnutls_srvconf_assign(ocsp_response_file_num);
     gnutls_srvconf_merge(ocsp_cache_time, MGS_TIMEOUT_UNSET);
     gnutls_srvconf_merge(ocsp_failure_timeout, MGS_TIMEOUT_UNSET);
+    gnutls_srvconf_merge(ocsp_fuzz_time, MGS_TIMEOUT_UNSET);
     gnutls_srvconf_merge(ocsp_socket_timeout, MGS_TIMEOUT_UNSET);
 
     gnutls_srvconf_assign(ca_list);
     gnutls_srvconf_assign(ca_list_size);
-    gnutls_srvconf_assign(cert_pgp);
-    gnutls_srvconf_assign(cert_crt_pgp);
-    gnutls_srvconf_assign(pgp_list);
     gnutls_srvconf_assign(certs);
     gnutls_srvconf_assign(anon_creds);
     gnutls_srvconf_assign(srp_creds);
@@ -1186,12 +989,6 @@ void *mgs_config_server_merge(apr_pool_t * p, void *BASE, void *ADD)
     gnutls_srvconf_assign(certs_x509_crt_chain);
     gnutls_srvconf_assign(certs_x509_chain_num);
 
-    /* how do these get transferred cleanly before the data from ADD
-     * goes away? */
-    gnutls_srvconf_assign(cert_cn);
-    for (i = 0; i < MAX_CERT_SAN; i++)
-	gnutls_srvconf_assign(cert_san[i]);
-
     return sc;
 }
 
