@@ -658,13 +658,7 @@ int mgs_hook_post_config(apr_pool_t *pconf,
         if (sc->enabled == GNUTLS_ENABLED_UNSET)
             sc->enabled = GNUTLS_ENABLED_FALSE;
         if (sc->tickets == GNUTLS_ENABLED_UNSET)
-        {
-            /* GnuTLS 3.6.4 introduced automatic master key rotation */
-            if (gnutls_check_version_numeric(3, 6, 4))
-                sc->tickets = GNUTLS_ENABLED_TRUE;
-            else
-                sc->tickets = GNUTLS_ENABLED_FALSE;
-        }
+            sc->tickets = GNUTLS_ENABLED_FALSE;
         if (sc->export_certificates_size < 0)
             sc->export_certificates_size = 0;
         if (sc->client_verify_mode == -1)
@@ -899,16 +893,19 @@ typedef struct {
 int check_server_aliases(vhost_cb_rec *x, server_rec * s, mgs_srvconf_rec *tsc)
 {
     apr_array_header_t *names;
-    int rv = 0;
     char ** name;
 
-    /* Check ServerName First! */
-    if (strcasecmp(x->sni_name, s->server_hostname) == 0) {
+    /* Check ServerName first */
+    if (strcasecmp(x->sni_name, s->server_hostname) == 0)
+    {
         // We have a match, save this server configuration
         x->sc = tsc;
-        rv = 1;
-        /* Check any ServerAlias directives */
-    } else if(s->names->nelts) {
+        return 1;
+    }
+
+    /* Check any ServerAlias directives */
+    if(s->names->nelts)
+    {
         names = s->names;
         name = (char **) names->elts;
         for (int i = 0; i < names->nelts; ++i)
@@ -917,13 +914,15 @@ int check_server_aliases(vhost_cb_rec *x, server_rec * s, mgs_srvconf_rec *tsc)
                 continue;
             if (strcasecmp(x->sni_name, name[i]) == 0)
             {
-                // We have a match, save this server configuration
                 x->sc = tsc;
-                rv = 1;
+                return 1;
             }
         }
-        /* ServerAlias directives may contain wildcards, check those last. */
-    } else if(s->wild_names->nelts) {
+    }
+
+    /* ServerAlias directives may contain wildcards, check those last. */
+    if(s->wild_names->nelts)
+    {
         names = s->wild_names;
         name = (char **) names->elts;
         for (int i = 0; i < names->nelts; ++i)
@@ -933,11 +932,11 @@ int check_server_aliases(vhost_cb_rec *x, server_rec * s, mgs_srvconf_rec *tsc)
             if (ap_strcasecmp_match(x->sni_name, name[i]) == 0)
             {
                 x->sc = tsc;
-                rv = 1;
+                return 1;
             }
         }
     }
-    return rv;
+    return 0;
 }
 
 static int vhost_cb(void *baton, conn_rec *conn, server_rec * s)
@@ -1155,6 +1154,11 @@ static void create_gnutls_handle(conn_rec * c)
             ap_log_cerror(APLOG_MARK, APLOG_ERR, err, c,
                           "gnutls_init for proxy connection failed: %s (%d)",
                           gnutls_strerror(err), err);
+        gnutls_handshake_set_hook_function(ctxt->session,
+                                           GNUTLS_HANDSHAKE_NEW_SESSION_TICKET,
+                                           GNUTLS_HOOK_POST,
+                                           mgs_proxy_got_ticket_func);
+        ctxt->proxy_ticket_key = mgs_proxy_ticket_id(ctxt, NULL);
     }
     else
     {
@@ -1165,6 +1169,11 @@ static void create_gnutls_handle(conn_rec * c)
             ap_log_cerror(APLOG_MARK, APLOG_ERR, err, c,
                           "gnutls_init for server side failed: %s (%d)",
                           gnutls_strerror(err), err);
+
+        /* Pre-handshake hook for early SNI parsing */
+        gnutls_handshake_set_hook_function(ctxt->session,
+                                           GNUTLS_HANDSHAKE_CLIENT_HELLO,
+                                           GNUTLS_HOOK_PRE, early_sni_hook);
     }
 
     /* Ensure TLS session resources are released when the connection
@@ -1177,11 +1186,6 @@ static void create_gnutls_handle(conn_rec * c)
         ap_log_cerror(APLOG_MARK, APLOG_ERR, err, c,
                       "gnutls_priority_set failed!");
 
-    /* Pre-handshake hook for early SNI parsing */
-    gnutls_handshake_set_hook_function(ctxt->session,
-                                       GNUTLS_HANDSHAKE_CLIENT_HELLO,
-                                       GNUTLS_HOOK_PRE, early_sni_hook);
-
     /* Post client hello hook (called after GnuTLS has parsed it) */
     gnutls_handshake_set_post_client_hello_function(ctxt->session,
             post_client_hello_hook);
@@ -1217,9 +1221,11 @@ static void create_gnutls_handle(conn_rec * c)
 
     /* Set pull, push & ptr functions */
     gnutls_transport_set_pull_function(ctxt->session,
-            mgs_transport_read);
+                                       mgs_transport_read);
+    gnutls_transport_set_pull_timeout_function(ctxt->session,
+                                               mgs_transport_read_ready);
     gnutls_transport_set_push_function(ctxt->session,
-            mgs_transport_write);
+                                       mgs_transport_write);
     gnutls_transport_set_ptr(ctxt->session, ctxt);
     /* Add IO filters */
     ctxt->input_filter = ap_add_input_filter(GNUTLS_INPUT_FILTER_NAME,
