@@ -212,18 +212,18 @@ static int mgs_create_ocsp_request(server_rec *s,
         return ret;
     }
 
-    ret = gnutls_ocsp_req_randomize_nonce(r);
-    if (ret != GNUTLS_E_SUCCESS)
-    {
-        ap_log_error(APLOG_MARK, APLOG_ERR, APR_EGENERAL, s,
-                     "OCSP nonce creation failed: %s (%d)",
-                     gnutls_strerror(ret), ret);
-        gnutls_ocsp_req_deinit(r);
-        return ret;
-    }
-
     if (nonce != NULL)
     {
+        ret = gnutls_ocsp_req_randomize_nonce(r);
+        if (ret != GNUTLS_E_SUCCESS)
+        {
+            ap_log_error(APLOG_MARK, APLOG_ERR, APR_EGENERAL, s,
+                         "OCSP nonce creation failed: %s (%d)",
+                         gnutls_strerror(ret), ret);
+            gnutls_ocsp_req_deinit(r);
+            return ret;
+        }
+
         ret = gnutls_ocsp_req_get_nonce(r, NULL, nonce);
         if (ret != GNUTLS_E_SUCCESS)
         {
@@ -705,7 +705,8 @@ static apr_status_t mgs_cache_ocsp_response(server_rec *s,
     else
     {
         gnutls_datum_t req;
-        int ret = mgs_create_ocsp_request(s, req_data, &req, &nonce);
+        int ret = mgs_create_ocsp_request(s, req_data, &req,
+                                          sc->ocsp_check_nonce ? &nonce : NULL);
         if (ret == GNUTLS_E_SUCCESS)
         {
             ap_log_error(APLOG_MARK, APLOG_TRACE2, APR_SUCCESS, s,
@@ -1269,7 +1270,7 @@ int mgs_ocsp_enable_stapling(apr_pool_t *pconf __attribute__((unused)),
     if (sc->ocsp_auto_refresh == GNUTLS_ENABLED_UNSET)
         sc->ocsp_auto_refresh = GNUTLS_ENABLED_TRUE;
     if (sc->ocsp_check_nonce == GNUTLS_ENABLED_UNSET)
-        sc->ocsp_check_nonce = GNUTLS_ENABLED_TRUE;
+        sc->ocsp_check_nonce = GNUTLS_ENABLED_FALSE;
     if (sc->ocsp_cache_time == MGS_TIMEOUT_UNSET)
         sc->ocsp_cache_time = apr_time_from_sec(MGS_OCSP_CACHE_TIMEOUT);
     if (sc->ocsp_failure_timeout == MGS_TIMEOUT_UNSET)
