@@ -139,14 +139,8 @@ class TestConnection(yaml.YAMLObject):
         self.actions = actions
         self.transport = Transports[transport.upper()]
         self.description = description
-        if host:
-            self.host = subst_env(host)
-        else:
-            self.host = os.environ.get('TEST_TARGET', 'localhost')
-        if port:
-            self.port = int(subst_env(port))
-        else:
-            self.port = int(os.environ.get('TEST_PORT', 8000))
+        self.host = host
+        self.port = port
 
     def __repr__(self):
         return (f'{self.__class__.__name__!s}'
@@ -158,6 +152,15 @@ class TestConnection(yaml.YAMLObject):
     def run(self, timeout=5.0, conn_log=None, response_log=None):
         """Set up an HTTP connection and run the configured actions."""
 
+        if self.host:
+            self.host = subst_env(self.host)
+        else:
+            self.host = os.environ.get('TEST_TARGET', 'localhost')
+        if self.port:
+            self.port = int(subst_env(self.port))
+        else:
+            self.port = int(os.environ.get('TEST_PORT', 8000))
+
         # note: "--logfile" option requires GnuTLS version >= 3.6.7
         command = ['gnutls-cli', '--logfile=/dev/stderr']
         for s in self.gnutls_params:
@@ -541,21 +544,20 @@ def subst_env(text):
 def run_test_conf(test_config, timeout=5.0, conn_log=None, response_log=None):
     """Load and run a test configuration.
 
-    The test_conf parameter must be a YAML file, defining one or more
-    TestConnections, to be run in order. The other three parameters
-    are forwarded to TestConnection.run().
+    The test_conf parameter must either a single TestConnection
+    object, or a list of such objects to be run in order. The other
+    three parameters are forwarded to TestConnection.run().
 
     """
     conns = None
 
-    config = yaml.load(test_config, Loader=yaml.Loader)
-    if type(config) is TestConnection:
-        conns = [config]
-    elif type(config) is list:
+    if type(test_config) is TestConnection:
+        conns = [test_config]
+    elif type(test_config) is list:
         # assume list elements are connections
-        conns = config
+        conns = test_config
     else:
-        raise TypeError(f'Unsupported configuration: {config!r}')
+        raise TypeError(f'Unsupported configuration: {test_config!r}')
     sys.stdout.flush()
 
     for i, test_conn in enumerate(conns):
